#include <unistd.h>
#include <stdarg.h>
#include <time.h>
#include <err.h>
#include <assert.h>
#include <png.h>
#include "img.h"
#include <stdio.h>
#include <stdlib.h>

#define GRAY50 (0x80)
#define GRAY99 (0xff)
#define GRAY00 (0x0)

time_t start_time, last_time;
FILE *log_file=NULL;

void log(char *fmt, ...)
{
  va_list ap;
  char buf[100];
  time_t now;

  va_start(ap, fmt);
  vsnprintf(buf, 100, fmt, ap);
  buf[99] = 0;

  now = time(NULL);
  printf("%3i[%3i]:%s\n", (int)(now - start_time), (int)(now - last_time), buf);
  if(log_file)
  {
    fprintf(
      log_file,
      "%3i[%3i]:%s\n",
      (int)(now - start_time),
      (int)(now - last_time),
      buf);
  }
  va_end(ap);
  last_time = now;
} /* log */

t_linefield*
linefield_new(int w, int h, int deltapercent)
{
  t_linefield *ret;
  int delta;
  int i;

  ret = (t_linefield*) malloc(sizeof(t_linefield));
  assert(ret);
 
  delta = (deltapercent *w) / 100;
  ret->w = w;
  ret->h = h;
  ret->delta = delta; 

  ret->mapx = delta * 2;
  ret->mapy = h + delta * 2;
  ret->mapxy = ret->mapx * ret->mapy;

  ret->map = malloc(sizeof(t_lfval) * ret->mapxy);
  assert(ret->map);

  ret->sum =0;
  ret->max =0;
  for(i=ret->mapxy; i--; ret->map[i]=0);

  return ret;

} /* linefield_new */

#define LF_MAP(lf, y, del) \
   ((lf)->map[(lf)->mapx * ((y) + (lf)->delta) + (del)+(lf)->delta])


t_lfval sumall = 0;

void
linefield_addweiht(t_linefield *lf, int x, int y, t_lfval vaha)
{
  int delta, dy;

  assert(x < lf->w);

  for(delta = - lf->delta + 1; delta  < lf->delta; delta++)
  {
    dy =  - ( delta * x ) / lf->w;
#if 1
    LF_MAP(lf, y + dy, delta) += vaha;
    sumall += vaha;
#else
   if( vaha > 16 )
   {
     LF_MAP(lf, y + dy, delta) ++;
     sumall ++;
   }   
#endif

 
    assert(delta + lf->delta < lf->mapy);
    assert(dy <  lf->delta);
    assert(dy > -lf->delta);

  }
} /* linefield_addweiht */



t_linefield*
linefield_olinuj(t_img *img)
{
  int x, y;
  unsigned short int **imgr = (unsigned short int**)img->r;
  t_linefield *lf;
  t_lfval val;

  assert(img->dep == 2 );

  lf = linefield_new(img->w, img->h, 40);

  for(y=img->h;y--;)
  for(x=img->w;x--;)
  {
    val = (t_lfval) (imgr[y][x]);
    linefield_addweiht(lf, x, y, val);
    lf->sum += val;
    assert(lf->sum >= 0);
  }

  return lf;
  
} /* linefield_new */

/*
 *  urci prah jako 3/4 maxima
 */
void
linefield_prah2(t_linefield *lf)
{
  int y,delta;

   t_lfval max = 0;

   for(y = -lf->delta;y<lf->h+lf->delta;y++)
   for(delta=-lf->delta;delta<lf->delta;delta++)
   {
     if( LF_MAP(lf, y, delta) > max )
     {
       max = LF_MAP(lf, y, delta);
     }
   }
   lf->prah = ( 3 *  max ) / 8;
} /* linefield_prah2 */
/*
 *  najde `num' nejtezsich prijimek
 */
void 
linefield_prah(t_linefield *lf, int num)
{
  int i, j;
  int *poc;
  int poc_siz;

  
  printf("linefield_prah...");
  fflush(stdout);

  
  poc_siz = 256 * lf->w;
  poc = (int*) malloc(sizeof(int) * poc_siz) ;
  assert(poc);

  for(i=poc_siz;i--;poc[i]=0);

  for(i = lf->mapxy;i--;)
  {
    assert(lf->map[i] < poc_siz);

    for(j = lf->map[i];j--;)
    {
      poc[j]++;
    }
  }

  /* najdi */
  for(j = poc_siz - 1 ; poc[j]<num && j; j--);

  lf->prah = j;

  printf("done[%i(%i)]\n", j, poc[j]);
} /* linefield_prah */


void
img_draw_top_lines(t_img *img, t_linefield *lf, t_lfval vaha)
{
  int y, yy, delta;
  int lines_drawed = 0;

  
  t_rgba col= { 0xff00, 0xff00, 0, 0};
  for(y = -lf->delta; y < lf->delta + lf->h;y++)
  {
    for(delta = -lf->delta + 1; delta < lf->delta; delta ++)
    {
      if( LF_MAP(lf, y, delta) >= vaha )
      { /* budeme kreslit tuto primiku */
        yy = y;
        assert( y + delta < lf->mapy );
        lines_drawed ++;
        img_line(img, 0, yy, lf->w - 1 , yy + delta, col);
        //printf("[%i, %i]: %Li\n", y, delta, LF_MAP(lf, y, delta));
      }
    }
  }
  printf("sum %Li\n", lf->sum);
  printf("max %Li\n", lf->max);
  printf("sumall %Li\n", sumall);
  printf("sumall/79 %Li\n", sumall/79);
  printf("vaha:%Li\n", vaha);
  printf("lines_drawed:%i\n", lines_drawed);
} /* img_draw_top_lines */


void
hrany_zvyraznit(t_img *s, t_img *d)
{
  int i,j;
  //unsigned short int **sr = (unsigned short int **) s->r;
  //unsigned short int **dr = (unsigned short int **) d->r;
  short int **sr = (short int **) s->r;
  short int **dr = (short int **) d->r;
  /* png_bytep * r = s->r; */

  assert(s->w==d->w);
  assert(s->h==d->h);
  assert(s->rb==d->rb);

  for(i=s->h;i--;dr[i][0] = dr[i][s->w-1] = GRAY50);
  for(j=s->w;j--;dr[0][j] = dr[s->h-1][j] = GRAY50);

  for(i=1; i < s->h - 1; i++)
  {
    for(j=1; j < s->w - 1 ; j++)
    {
#if 1
      dr[i][j] =  (sr[i-1][j-1] * 2 + sr[i-1][j] + sr[i-1][j+1] * 2 -
                  sr[i+1][j-1] * 2 - sr[i+1][j] - sr[i+1][j+1]*2)/10 + GRAY50;
     assert(dr[i][j] < 260);
     assert(dr[i][j] > 0);
#else
 
      dr[i][j] =  sr[i-1][j] + sr[i+1][j] +
                  sr[i][j+1] + sr[i][j+1] - 4* sr[i][j] + GRAY50;
#endif
    }
  } 
  
} /* hrany_zvyraznit */

void
uber_sedou(t_img *img)
{

  int i;
  unsigned short int * imgb = (unsigned short int*) img->b;

  assert(img->dep == 2);
  

  for(i=img->h*img->w; i--;)
  {
    if(imgb[i]<=GRAY50)
    {
      imgb[i]=GRAY00;
    }
    else
    {
      //imgb[i]-=GRAY50; // imgb[i]-GRAY50;
      imgb[i]= (imgb[i] - GRAY50)*2;
    }
  }
} /* uber_sedou */ 

/*
 *  dokresli puntiky na mista primek
 */
void
lf_img_puntiky(t_linefield *lf, t_li_set *lis,  t_img *img)
{
  int i;
  int y, d;
  int z;
  t_rgba col={0xffff, 0, 0, 0};

  for(i = lis->top;i--;)
  {
    y = lis->m[i].fsty;
    z = lis->m[i].fstz - y;
//printf("lf_img_puntiku (%i, %i)\n", y, z);
    y += lf->delta;
    z += lf->delta;

//printf("lf_img_puntiku (%i, %i)\n", y, z);

    img_line(img, z + 5, y - 5, z - 5, y + 5,  col);
    img_line(img, z - 5, y - 5, z + 5, y + 5,  col);
  }

} /* lf_img_puntiky */

void
usage()
{
  printf(
   ". -oOut.png -OOut2.png -HHo.png In.png\n");
} /* usage */


/*******
 *  nalezeni opravdovych primek
 */
#define LI_ARRAY_MAX 300

//t_lineinterval li_array[LI_ARRAY_MAX];

//int li_array_top =0;

/*
 * line interval 
 */
t_li_set*
li_set_new()
{
  t_li_set *lis;

  lis = (t_li_set*)malloc(sizeof(t_li_set));
  assert(lis);

  lis->top = 0;
  lis->siz = LI_ARRAY_MAX;

  lis->m = (t_lineinterval*) malloc(sizeof(t_lineinterval)*lis->siz);
  assert(lis->m);

  return lis;
} /* li_set_new */

void
lf_usek_one(t_lineinterval *li, t_img *img0, t_img *img,  int mezery,  int prah)
{

  int x, y, col;
  unsigned short int **imgr = (unsigned short int**) img->r;
  int d, w, y0, mx_start, mx_len, ted_start, ted_len;
  int muzu_vynechat;
  t_rgba rgba = { 0xffff, 0xffff, 0, 0};

  assert(img->dep == 2);

  d = li->fstz - li->fsty;
  w = img->w -1;
  y0 = li->fsty;

  mx_start = 0;
  mx_len = 0;

  ted_start = 0;
  ted_len = 0;

  muzu_vynechat = 0;
  
  for(x = 0 ; x<=w; x++,ted_len++)
  {
    y = (d * x ) / w + y0;

    if(y<0||y>=img->h)
      col =0;
    else 
       col = imgr[y][x] >0;

//printf("%c", col?'*':' ');
    if(col)
    {
      //img_putpixel(img0, x, y, rgba);
      if(muzu_vynechat <= 0)
      { /* zacatek noveho useku  (protoze velka dira */
        ted_start = x;
        ted_len = 0;
      }
      muzu_vynechat = mezery;
    }
    else 
    {
      if(muzu_vynechat == mezery )
      { /* prvni nula po dlouhe dobe */
        if(ted_len > mx_len)
        {
          mx_start = ted_start;
          mx_len = ted_len;
        }
      }
      muzu_vynechat--;
    }
  } /* for */
  if(ted_len > mx_len)
  {
    mx_start = ted_start;
    mx_len = ted_len;
  }

  li->xmin = mx_start;
  li->xmax = mx_start + mx_len - 1 ;
//printf("lf_usek<%i,%i>\n", li->xmin, li->xmax);
  
} /* lf_usek_one */

void
lf_usek(t_li_set *lis, t_img *img0, t_img *img,  int mezery,  int prah)
{
  int i;

  for(i=lis->top;i--;)
  {
    lf_usek_one(&lis->m[i], img0, img, mezery, prah);
  }
} /* lf_usek */

/*
 * kresli jen tam kde to je, bere prumernou primku
 */
void
li_draw_usek(t_li_set *lis, t_img *img)
{
  t_rgba rgba={0xffff, 0, 0, 0x0};
  int li;
  int y, z;
  int xmin, xmax, ymin, ymax;
  int w = (img->is_rot?img->h:img->w) -1;
  int y0, d;

  for(li=lis->top; li--; )
  {
    y0 = lis->m[li].fsty;
    d = lis->m[li].fstz - y0 ;

    xmax = lis->m[li].xmax;
    xmin = lis->m[li].xmin;

    ymin = (d * xmin ) / w + y0;
    ymax = (d * xmax ) / w + y0;
    //img_line_jenkdeje(img, img2, 0, xmin,  y, img->w, xmax, rgba);
    //img_line(img, 0,  y, img->w - 1, z, rgba);
    img_line(img, xmin,  ymin, xmax, ymax, rgba);
  }

} /* li_draw */

/*
 *  prida novy line interval (siroky 0)
 */
void
li_addli(t_li_set *lis, int y, int d)
{
  if( lis->top >= LI_ARRAY_MAX )
  {
     err(1, "li_addli: li_array_top overflow");
  }

  lis->m[lis->top].fsty = 
  lis->m[lis->top].ymin = 
  lis->m[lis->top].ymax = y;

  lis->m[lis->top].fstz = 
  lis->m[lis->top].zmin = 
  lis->m[lis->top].zmax = y + d;

  lis->top ++;
} /* li_addli */

/*
 * otestne zdali uvedeny interval ma prunik s danou linou 
 * r: !0 jesli to pridal
 */
void
li_addline(t_li_set *lis, int y, int d)
{
  int li;
  int ymx, ymn, zmx, zmn;


  for(li=lis->top; li--;)
  {
  
    ymn = lis->m[li].ymin;
    ymx = lis->m[li].ymax;
    
    zmn = lis->m[li].zmin;
    zmx = lis->m[li].zmax;

    if( y > ymx + 1 && y + d > zmx + 1 )
    { /* cela primka je pod */
      continue;
    }
  
    if( y < ymn - 1  && y + d < zmn - 1)
    { /* cela primka je nad */
      continue;
    }
  
    /* pridame tuhle caru */
    if( y  < ymn )
    {
      lis->m[li].ymin = y;
    }
    else if( y > ymx )
    {
      lis->m[li].ymax = y;
    }
  
    /* pridame tuhle caru */
    if( y + d  < zmn  )
    {
      lis->m[li].zmin = y + d;
    }
    else if( y + d > zmx )
    {
      lis->m[li].zmax = y + d;
    }
    return;
  }
  li_addli(lis, y,d);
} /* li_addline */


t_lfval 
lf_val_at(t_linefield * lf, int y, int d)
{
  if( y < -lf->delta ||
      y >= lf->h + lf->delta ||
      d < -lf->delta ||
      d >= lf->delta)
   {
     printf("lf_val_at:pres okraj(y%i, d%i)\n", y, d);
     return 0;
   }
   return LF_MAP(lf, y, d);
} /* lf_val_at */

/*
 * najdi lokalni maximum v okoli y z
 * a posun prislusne py, pz 
 * r: !0 bylo posunuto
 */
int
li_findmx_one(t_linefield *lf, int *py, int *pd)
{

  t_lfval v[9];
  int y = *py;
  int d = *pd;
  int i;
  int mx, mxi;
  int dy, dd;


  v[0] = lf_val_at(lf, y - 1, d - 1);
  v[1] = lf_val_at(lf, y - 1, d    );
  v[2] = lf_val_at(lf, y - 1, d + 1);

  v[3] = lf_val_at(lf, y    , d - 1);
  v[4] = lf_val_at(lf, y    , d    );
  v[5] = lf_val_at(lf, y    , d + 1);

  v[6] = lf_val_at(lf, y + 1, d - 1);
  v[7] = lf_val_at(lf, y + 1, d    );
  v[8] = lf_val_at(lf, y + 1, d + 1);
  
  
  /* najdi mzximum */
  mxi = 8;
  mx = v[8];

  for(i=8;i--;)
  {
    if(mx < v[i])
    {
      mxi = i;
      mx = v[i];
    }
  }

  /* spocti posun */
  dy = (mxi / 3) - 1;
  dd = (mxi % 3) - 1;

  *py += dy;
  *pd += dd;

  /* pokud sem se neposunul ses mekej */
  return dy || dd;

} /* li_findmx_one */

void
li_findmx(t_li_set *lis, t_linefield *lf)
{
  int i;
  int y, d;

  for(i = lis->top;i--;)
  {
    y = lis->m[i].fsty;
    d = lis->m[i].fstz - y;

    assert(lf_val_at(lf, y, d) > 0);

    while(li_findmx_one(lf,&y,&d));

    lis->m[i].fsty = y;
    lis->m[i].fstz = y + d;
  }
} /* li_findmx */

/*
 *  najde cary ktere se protinaji a spoji je do jednine 
 */
void
li_eqline(t_li_set *lis, t_linefield *lf, t_lfval vaha)
{
  int y, delta;

  for(y = lf->delta + lf->h -1 ;  y > -lf->delta +1;y--)
  {
    for(delta= lf->delta; delta > -lf->delta;delta--)
    {
      if( LF_MAP(lf, y, delta) > vaha )
      {
        if(lf_val_at(lf, y, delta) < vaha)
        {
          printf("y%i delta%i lf_val_at%Li LF_MAP%Li\n",
            y, delta,
            lf_val_at(lf, y, delta),
            LF_MAP(lf, y, delta));
          exit(1);
        };
        li_addline(lis, y, delta);
      }
    }
  }
  log("li_eqpine:lis->top:%i", lis->top);
} /* li_eqline */
/*
 * kresli jen tam kde to je, bere prumernou primku
 */
void
li_draw(t_li_set *lis, t_img *img, t_img *img2)
{
  t_rgba rgba={0, 0xffff, 0, 0x10};
  int li;
  int y, z;

  for(li=lis->top; li--; )
  {
    y = (lis->m[li].ymin + lis->m[li].ymax)/2;
    z = (lis->m[li].zmin + lis->m[li].zmax)/2;
    img_line_jenkdeje(img, img2, 0, 0,  y, img->w - 1 , z, rgba);
  }

} /* li_draw */

/*
 * kresli jen tam kde to je, bere primku fst
 */
void
li_draw_fst(t_li_set *lis, t_img *img, t_img *img2)
{
  t_rgba rgba={0, 0xffff, 0, 0x10};
  int li;
  int y, z;

  for(li=lis->top; li--; )
  {
    y = lis->m[li].fsty;
    z = lis->m[li].fstz;
    img_line_jenkdeje(img, img2, 0, 0,  y, img->w - 1, z, rgba);
  }

} /* li_draw */

void
li_drawrot(t_li_set *lis, t_img *img)
{
  t_rgba rgba={0, 0, 0xffff, 0x10};
  int li;
  int y, z;

  for(li=lis->top; li--; )
  {
    y = (lis->m[li].ymin + lis->m[li].ymax)/2;
    z = (lis->m[li].zmin + lis->m[li].zmax)/2;
    img_line(img,  y, 0, z, img->h - 1, rgba);
  }

} /* li_draw */

char *name_in = "in.png";
char *name_out = "out.png";
char *name_out2 = "out2.png";
char *log_filename = NULL;
char *hough_filename = "ho-le.png";
char *hough_up_filename = "ho-up.png";

void
parse_args(int argc, char *argv[])
{
  int c;
  img_load_font("iso02.font");

  while( -1 != (c = getopt(argc, argv, "G:H:l:O:o:")))
  { 
    switch(c)
    {
    case 'o': name_out = optarg; break;
    case 'O': name_out2 = optarg; break;
    case 'l': log_filename = optarg; break;
    case 'H': hough_filename = optarg; break;
    case 'G': hough_up_filename = optarg; break;
    }
  }

  argv += optind;
  argc -= optind;

  if( argc > 1  )
  {
    usage();
    exit(0);
  }

  if(log_filename)
  {
    if( NULL == (log_file = fopen(log_filename, "at")))
    {
      err(1, "cannt open %s", log_filename);
    }
  }

  if( argc == 1)
  {
    name_in = argv[0];
  }


}/* parse_args */


int
main(int argc, char *argv[])
{
  //t_rgba col= { 0xff00, 0xff00, 0, 0x0};
  //int i;
  t_lfval prah, prah2;
  t_img *img, *img2, *img3, *img4;
  t_img *imgrot, *imgrot2, *imgrot4;
  t_li_set *lis, *lis2;
  t_img *imglf, *imglfcol;

  t_linefield *lf, *lf2;

  lis = li_set_new();
  lis2 = li_set_new();

  last_time =
  start_time = time(NULL);

  parse_args(argc, argv);

  img = img_png_read(name_in);

  if ( img == NULL )
  {
    printf("cann't read_png(%s)", name_in);
  }


  log("start(%s)", name_in);
  img2 = img_new(img->w, img->h, 2);

#define UP 1
#define HO 1

#if UP
  imgrot = img_rot(img);
  imgrot2 = img_new(imgrot->w, imgrot->h, 2);
#endif


  log("bf zvyrazenene hrany");
  hrany_zvyraznit(img, img2);

#if UP
  hrany_zvyraznit(imgrot, imgrot2);
#endif
  log("af zvyrazenene hrany");

  uber_sedou(img2);
#if UP
  uber_sedou(imgrot2);
#endif

  img3 = img_gray2rgb(img2);
  img4 = img_gray2rgb(img);

  //imgrot4 = img_gray2rgb(imgrot);

  log("bf hough (olinuj)");
  lf = linefield_olinuj(img2);
#if UP
  log("bf hough (olinuj) rot");
  lf2 = linefield_olinuj(imgrot2);
#endif
  log("af linefield_olinuj");

  log("bf prah");
  linefield_prah(lf, 300);
  //linefield_prah2(lf);
#if UP
  linefield_prah(lf2, 300);
  //linefield_prah2(lf2);
#endif
  log("af prah");

  //linefield_prah2(lf);
  //linefield_prah(lf, 300);
  prah = lf->prah;
#if UP
  prah2 =  lf2->prah;
#endif
  
  //img_draw_top_lines(img3, lf2, prah);
  
  log("bf lf_eqline");
  li_eqline(lis, lf, prah);
  li_findmx(lis, lf);

  log("bf lf_eqline rot");
#if UP
  li_eqline(lis2,lf2, prah2);
  li_findmx(lis2, lf2);
#endif
  log("af lf_eqline");

  log("lis->top:%i", lis->top);
#if UP
  log("lis2->top:%i", lis2->top);
#endif

  log("bf li_draw_fst");
  li_draw_fst(lis, img4, img2);
#if UP
  img_is_rot(img4);
  li_draw_fst(lis2, img4, imgrot2);
  img_is_rot(img4);
#endif
  log("af li_draw_fst");

  log("bf lf_usek");
  lf_usek(lis, img4, img2, 3/*mezery*/, 0/*prah*/);
  li_draw_usek(lis, img4);

#if 0
  log("bf lf_usek rot");
  img_is_rot(img4);
  lf_usek(lis2, img4, imgrot2, 3/*mezery*/, 0/*prah*/);
  li_draw_usek(lis, img4);
  img_is_rot(img4);
  log("af lf_usek");
#endif

  //li_draw(img3);
  log("af li_draw");

#if HO
  log("bf lf_to_img");
  imglf = lf_to_img(lf);
  imglfcol = img_gray2rgb(imglf);
  lf_img_puntiky(lf, lis,  imglfcol);
  log("bf write hu");
  if( -1 == img_png_write( hough_filename, imglfcol) )
  {
    printf("cann't write_png(%s)", hough_filename);
  }
#if UP
  log("bf lf_to_img2");
  imglf = lf_to_img(lf2);
  imglfcol = img_gray2rgb(imglf);
  lf_img_puntiky(lf2, lis2, imglfcol);
  log("bf write hu");
  if( -1 == img_png_write(  hough_up_filename, imglfcol) )
  {
    printf("cann't write_png(%s)", hough_up_filename);
  }
#endif UP
#endif
  log("af lf_to_img");




  log("lis->top:%i", lis->top);

  log("bf writing");
  if( -1 == img_png_write( name_out, img4) )
  {
    printf("cann't write_png(%s)", name_out);
  }
#if 0 
  if( -1 == img_png_write( name_out2, img) )
  {
    printf("cann't write_png(%s)", name_out2);
  }
#endif
  log("end\n");
  return 0;
} /* main */

